# Forecasting/autoforecast/main.py
import time

import numpy as np
from tqdm import tqdm

from autoforecast.metrics import get_metrics
from autoforecast.models.models import get_dict_models
from autoforecast.src.utils.logger import LOG


class AutoForecast:
    def __init__(self, metrics=None):
        """
        metrics: str, metric to optimize
        """
        self.dict_models = get_dict_models()
        self.model = None

    def fit(self, X_train, y_train):
        # given X_train, y_train find the best model
        cutoff = int(len(X_train) * 0.7)
        X_train_ = X_train[:cutoff]
        y_train_ = y_train[:cutoff]
        X_test_ = X_train[cutoff:]
        y_test_ = y_train[cutoff:]
        res_auto_forecast = self.run_auto_forecast(
            X_train=X_train_,
            y_train=y_train_,
            X_test=X_test_,
            y_test=y_test_,
            verbose=False,
            max_time_in_sec=360,
        )
        self.model = res_auto_forecast["best_model"]["model"]
        best_param = res_auto_forecast["best_model"]["best_param"]
        self.model.fit(X_train, y_train, **best_param)

    def predict(self, X_test):
        y_pred = self.model.predict(X_test)
        return y_pred

    # custom AutoForecast
    def run_auto_forecast(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_test: np.ndarray,
        y_test: np.ndarray,
        verbose: bool = False,
        max_time_in_sec: int = 360,
    ):
        """
        Args:
            :X_train: np.ndarray
            :y_train: np.ndarray
            :X_test: np.ndarray
            :y_test: np.ndarray
            :verbose: bool = False
            :max_time_in_sec: int = 360
        """
        LOG.debug("AutoForecast starting...")
        start = time.time()
        dict_metrics = {}
        dict_pred = {}
        dict_trained_model = {}
        for model_str, model in tqdm(self.dict_models.items()):
            LOG.debug(model_str)
            dict_trained_model[model_str] = {}
            dict_trained_model[model_str]["model"] = model
            best_param = {}
            try:
                LOG.debug("start optimize hyperparameters")
                res_opt, best_param = model.optimize(
                    X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test
                )
                LOG.debug(res_opt)
                LOG.debug(best_param)
            except AttributeError as err:
                LOG.debug(f"except: {err}")
            model.fit(X_train, y_train, **best_param)
            dict_trained_model[model_str]["best_param"] = best_param
            y_pred = model.predict(X_test)
            dict_pred[model_str] = y_pred
            if verbose:
                LOG.debug(model_str)
                LOG.debug(f"pred={y_pred}")
            metrics = get_metrics(y_test=y_test, y_pred=y_pred)
            dict_metrics[model_str] = metrics
            if start - time.time() >= max_time_in_sec:
                break
        # log best models
        # SMAPE
        LOG.debug("Best models according to SMAPE metrics:")
        dict_metrics_smape = {k: v["smape"] for k, v in dict_metrics.items()}
        # sorted dict
        dict_metrics_smape = dict(sorted(dict_metrics_smape.items(), key=lambda item: item[1]))
        LOG.debug(dict_metrics_smape)
        # RMSE
        LOG.debug("Best models according to RMSE metrics:")
        dict_metrics_rmse = {k: v["rmse"] for k, v in dict_metrics.items()}
        # sorted dict
        dict_metrics_rmse = dict(sorted(dict_metrics_rmse.items(), key=lambda item: item[1]))
        LOG.debug(dict_metrics_rmse)
        LOG.debug("Best models according to MAPE metrics:")
        # MAPE
        dict_metrics_mape = {k: v["mape"] for k, v in dict_metrics.items()}
        # sorted dict
        dict_metrics_mape = dict(sorted(dict_metrics_mape.items(), key=lambda item: item[1]))
        LOG.debug(dict_metrics_mape)
        LOG.debug(f"AutoForecast done in {round(time.time()-start, 2)}s.")
        best_model_str = list(dict_metrics_smape.keys())[0]
        return {
            "dict_metrics": dict_metrics,
            "dict_pred": dict_pred,
            "best_model": dict_trained_model[best_model_str],
        }



import os

from dotenv import load_dotenv

load_dotenv()


class Logger:
    def __init__(self):
        pass

    def debug(self, message: str) -> None:
        if self.is_verbose():
            print("[DEBUG]{}".format(self.clean_message(message)))

    def info(self, message: str) -> None:
        if self.is_verbose():
            print("[INFO]{}".format(self.clean_message(message)))

    def warning(self, message: str) -> None:
        print("[WARNING]{}".format(self.clean_message(message)))

    def error(self, message: str) -> None:
        print("[ERROR]{}".format(self.clean_message(message)))

    def critical(self, message: str) -> None:
        print("[CRITICAL]{}".format(self.clean_message(message)))

    def is_verbose(self) -> bool:
        if os.getenv("DEBUG") is not None:
            return True

        return False

    def increase_verbosity(self):
        self.lowLogVerbosity = []

    def clean_message(self, message: str) -> str:
        return message


LOG = Logger()



# Forecasting/autoforecast/main.py
from autoforecast.automl import AutoForecast
from autoforecast.datasets.import_bitcoin_price import get_price_for_last_n_days
from autoforecast.src.utils.logger import LOG


def run(verbose: bool = False):
    LOG.debug("autoforecast_bitcoin.run() starting...")
    # settings
    # lists of features name
    list_cat_feat = ["timestamp"]
    # lists of features name tokenized
    list_num_feat = []

    df_price = get_price_for_last_n_days(n=900, type="spot", currency_pair="BTC-USD")
    df_price = df_price.rename(columns={"price": "target"})
    LOG.info(df_price)

    ind_cutoff = int(df_price.shape[0] * 0.8)
    train = df_price.iloc[:ind_cutoff]
    test = df_price.iloc[ind_cutoff:]
    LOG.info(f"{train.shape} {test.shape}")

    cols = list_cat_feat + list_num_feat
    X_train = train[cols].values
    y_train = train["target"].values
    X_test = test[cols].values
    y_test = test["target"].values
    model = AutoForecast()
    LOG.debug("Autoforecast() model fitting...")
    model.fit(X_train=X_train, y_train=y_train)
    LOG.debug("Autoforecast() model predicting...")
    y_pred = model.predict(X_test=X_test)
    LOG.debug(f"shapes={X_train.shape}{X_test.shape}{y_train.shape}{y_test.shape}")
    LOG.debug(f"y_pred={y_pred}")
    LOG.debug(f"y_test={y_test}")
    LOG.debug("autoforecast_bitcoin.run() ending...")



"""
encoding categorical features
"""


def encode_one_col(data, col: str):
    map_col_to_col_id = {col: col_id for col_id, col in enumerate(data[col].unique())}
    data[f"{col}_token"] = data[col].map(map_col_to_col_id)
    return data, map_col_to_col_id


def encode(data, list_col):
    dict_map = {}
    for col in list_col:
        data, map_col_to_col_id = encode_one_col(data, col)
        dict_map[col] = map_col_to_col_id
    return data, dict_map


def split_date(df, date_col):
    """
    split YYYY-mm-dd date into year, month, day
    """
    list_date = df[date_col].tolist()
    list_year = [int(date[:4]) for date in list_date]
    list_month = [int(date[5:7]) for date in list_date]
    list_day = [int(date[8:]) for date in list_date]
    df["year"] = list_year
    df["month"] = list_month
    df["day"] = list_day
    return df


def run(df, list_cat_feat, date_col):
    if date_col:
        df = split_date(df, date_col)
        list_cat_feat += ["year", "month", "day"]
    df, dict_map = encode(data=df, list_col=list_cat_feat)
    list_cat_feat_token = [f"{col}_token" for col in list_cat_feat]
    return df[list_cat_feat_token].values



# Forecasting/autoforecast/src/models/time_series_models.py
import datetime

import fbprophet
import numpy as np
import pandas as pd
import statsmodels.api as sm
from autoforecast import metrics
from autoforecast.configs.configspace.time_series_space import prophet_space, prophet_x0
from autoforecast.models.hyperparameters import HyperparametersTuner


class ARMA:
    def __init__(self, order=None, period=None):
        self.order = order

    def fit(self, X_train, y_train):
        self.y_train = y_train
        aic_matrix = self.build_aic(
            y_train=self.y_train, X_train=X_train, p_max=6, q_max=6, p_min=0, q_min=0
        )
        self.order = self.best_order(aic_matrix)

    def predict(self, X_test, *arg):
        y_pred = self.predict_arma(y_train=self.y_train, nforecast=len(X_test), order=self.order)
        return y_pred

    @staticmethod
    def build_aic(y_train: np.ndarray, X_train: np.ndarray, p_max=6, q_max=6, p_min=0, q_min=0):
        aic_full = pd.DataFrame(np.zeros((6, 6), dtype=float))
        # Iterate over all ARMA(p,q) models with p,q in [0,6]
        for p in range(6):
            for q in range(6):
                if p == 0 and q == 0:
                    continue
                # Estimate the model with no missing datapoints
                mod = sm.tsa.statespace.SARIMAX(
                    y_train, exog=X_train, order=(p, 0, q), enforce_invertibility=False
                )
                try:
                    res = mod.fit(disp=False, maxiter=200)
                    aic_full.iloc[p, q] = res.aic
                except Exception as e:
                    print(e)
                    aic_full.iloc[p, q] = np.nan
        return aic_full

    @staticmethod
    def best_order(aic_full: pd.DataFrame):
        aic_full.iloc[0, 0] = 10.0 ** 99
        # min column
        min_col_name = aic_full.min().idxmin()
        # min column index if needed
        min_col_idx = aic_full.columns.get_loc(min_col_name)
        # min row index
        min_row_idx = aic_full[min_col_name].idxmin()
        return (min_row_idx, 0, min_col_idx)

    @staticmethod
    def predict_arma(y_train: np.ndarray, nforecast: int, order=(1, 0, 1)):
        mod = sm.tsa.statespace.SARIMAX(y_train, order=order)
        res = mod.fit(disp=False)
        # In-sample one-step-ahead predictions, and out-of-sample forecasts
        predict = res.get_prediction(end=mod.nobs + nforecast)
        y_pred = predict.predicted_mean[-nforecast:]
        return np.array(y_pred)


class Prophet:
    def __init__(self):
        self.model = None
        self.period = None

    def fit(self, X_train, y_train, **params):
        numdays = len(X_train)
        base = datetime.datetime.today()
        date_list = [base - datetime.timedelta(days=x) for x in range(numdays)]
        df = pd.DataFrame({"ds": date_list, "y": y_train})
        self.model = fbprophet.Prophet(**params)
        self.model.fit(df)

    def predict(self, X_test):
        self.period = len(X_test)
        future = self.model.make_future_dataframe(periods=self.period, freq="M")
        forecast = self.model.predict(future)
        y_pred = forecast.yhat[-self.period :]
        return y_pred.values

    def optimize(self, X_train, X_test, y_train, y_test):
        return HyperparametersTuner(
            model=Prophet,
            search_space=prophet_space,
            x0=prophet_x0,
            metric=metrics.smape_score,
            X_train=X_train,
            X_test=X_test,
            y_train=y_train,
            y_test=y_test,
        )()



# src/utils/utils.py
import pandas as pd
from sklearn.preprocessing import MinMaxScaler


def encode(data, col="bank"):
    map_col_to_col_id = {col: col_id for col_id, col in enumerate(data[col].unique())}
    data[f"{col}_token"] = data[col].map(map_col_to_col_id)
    return data, map_col_to_col_id


def split_n_last_timestep(df, n=12):
    train, test = df.iloc[:-n], df.iloc[-n:]
    return train, test


def scale_transform(train, test=None):
    scaler = MinMaxScaler()
    scaler.fit(train)
    train = scaler.transform(train)
    if test is not None:
        test = scaler.transform(test)
    return train, test, scaler


def prepare_for_training(df, cols_to_drop=["heading_id", "company_id"]):
    df = df.drop(cols_to_drop, axis=1)
    df.date = pd.to_datetime(df.date)
    df = df.set_index("date")
    if "month" not in cols_to_drop:
        df["month"] = df["month"].astype("float64")
    return df


def train_test_split(df):
    train, test = split_n_last_timestep(df, n=12)
    return train, test



import plotly.graph_objects as go

from autoforecast.src.utils.logger import LOG


def plot_scatter(y_true, y_pred, x):
    fig = go.Figure(data=go.Scatter(x=x, y=y_true, name="true"))
    # fig.add_trace(go.Scatter(x=x, y=y_true, name='true'))
    fig.add_trace(go.Scatter(x=x, y=y_pred, name="pred"))
    title = "my title"
    LOG.debug(title)
    fig.show()



# Forecasting/autoforecast/src/models/keras_models.py
import numpy as np
import tensorflow as tf
from autoforecast import metrics
from autoforecast.configs.configspace.neural_net_space import (
    base_keras_space,
    base_keras_x0,
    lstm_keras_space,
    lstm_keras_x0,
)
from autoforecast.models.hyperparameters import HyperparametersTuner
from keras import Model
from keras.layers import LSTM, Dense, Input


class BaseKeras:
    def __init__(self, n_input=12, n_features=1):
        self.n_input = n_input
        self.n_features = 1

    def fit(self, X_train, y_train, **params):
        self.X_train = np.array(
            [y_train[i : i + self.n_input] for i in range(len(y_train) - self.n_input)]
        )
        self.y_train = np.array(
            [y_train[i + 1 : i + self.n_input + 1] for i in range(len(y_train) - self.n_input)]
        )

        self.model = self.keras_model(self.n_input, self.n_features, **params)
        self.model.fit(self.X_train, self.y_train, epochs=10, validation_split=0.1, verbose=0)

    def predict(self, X_test, *args):
        self.n_input = len(X_test)
        pred_list = self.predict_by_batch(self.model, self.X_train, self.n_input, self.n_features)
        return pred_list

    @staticmethod
    def keras_model(n_input, n_features, hidden_size_dense=3):
        inputs = Input(shape=(n_input, n_features))
        x = Dense(hidden_size_dense, activation="relu")(inputs)
        outputs = Dense(units=1)(x)
        model = Model(inputs=inputs, outputs=outputs)
        loss = tf.keras.losses.MeanAbsolutePercentageError()
        optimizer = tf.keras.optimizers.Adam()
        model.compile(loss=loss, optimizer=optimizer)
        return model

    @staticmethod
    def predict_by_batch(model, X_train, n_input, n_features):
        pred_list = []
        batch = X_train[-1].reshape(-1, 1)
        for i in range(n_input):
            pred_list.append(model.predict(batch)[0])
            batch = np.append(batch[1:], pred_list[i].reshape(-1, 1), axis=0)
        pred_list = np.concatenate([i for i in pred_list]).reshape(-1, 1)
        return np.concatenate(pred_list)

    def optimize(self, X_train, X_test, y_train, y_test):
        return HyperparametersTuner(
            model=BaseKeras,
            search_space=base_keras_space,
            x0=base_keras_x0,
            metric=metrics.smape_score,
            X_train=X_train,
            X_test=X_test,
            y_train=y_train,
            y_test=y_test,
        )()


class LSTMKeras(BaseKeras):
    def __init__(self):
        return super().__init__()

    def fit(self, X_train, y_train, **params):
        return super().fit(X_train, y_train, **params)

    def predict(self, *args):
        return super().predict(*args)

    @staticmethod
    def keras_model(n_input, n_features, hidden_size_lstm=3):
        inputs = Input(shape=(n_input, n_features))
        x = LSTM(hidden_size_lstm, activation="relu")(inputs)
        outputs = Dense(units=1)(x)
        model = Model(inputs=inputs, outputs=outputs)
        loss = tf.keras.losses.MeanAbsolutePercentageError()
        optimizer = tf.keras.optimizers.Adam()
        model.compile(loss=loss, optimizer=optimizer)
        return model

    def optimize(self, X_train, X_test, y_train, y_test):
        return HyperparametersTuner(
            model=LSTMKeras,
            search_space=lstm_keras_space,
            x0=lstm_keras_x0,
            metric=metrics.smape_score,
            X_train=X_train,
            X_test=X_test,
            y_train=y_train,
            y_test=y_test,
        )()



import torch
import torch.nn as nn

train_data_normalized = list(map(float, range(50)))

train_data_normalized = torch.FloatTensor(train_data_normalized).view(-1)


def create_inout_sequences(input_data, tw):
    inout_seq = []
    L = len(input_data)
    for i in range(L - tw):
        train_seq = input_data[i : i + tw]
        train_label = input_data[i + tw : i + tw + 1]
        inout_seq.append((train_seq, train_label))
    return inout_seq


train_window = 12
train_inout_seq = create_inout_sequences(train_data_normalized, train_window)


class LSTM(nn.Module):
    def __init__(self, input_size=1, hidden_layer_size=100, output_size=1):
        super().__init__()
        self.hidden_layer_size = hidden_layer_size

        self.lstm = nn.LSTM(input_size, hidden_layer_size)

        self.linear = nn.Linear(hidden_layer_size, output_size)

        self.hidden_cell = (
            torch.zeros(1, 1, self.hidden_layer_size),
            torch.zeros(1, 1, self.hidden_layer_size),
        )

    def forward(self, input_seq):
        lstm_out, self.hidden_cell = self.lstm(
            input_seq.view(len(input_seq), 1, -1), self.hidden_cell
        )
        predictions = self.linear(lstm_out.view(len(input_seq), -1))
        return predictions[-1]


def train_model():
    model = LSTM()
    loss_function = nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)

    epochs = 150

    for i in range(epochs):
        for seq, labels in train_inout_seq:
            optimizer.zero_grad()
            model.hidden_cell = (
                torch.zeros(1, 1, model.hidden_layer_size),
                torch.zeros(1, 1, model.hidden_layer_size),
            )

            y_pred = model(seq)

            single_loss = loss_function(y_pred, labels)
            single_loss.backward()
            optimizer.step()

        if i % 25 == 1:
            print(f"epoch: {i:3} loss: {single_loss.item():10.8f}")

    print(f"epoch: {i:3} loss: {single_loss.item():10.10f}")

    test_inputs = train_data_normalized[-train_window:].tolist()
    print(test_inputs)



import xgboost as xgb
from autoforecast.models.naive import BaselineLastValue, BaselineMean, BaselineMedian
from autoforecast.models.neural_net import BaseKeras, LSTMKeras
from autoforecast.models.time_series import ARMA, Prophet
from sklearn.ensemble import AdaBoostRegressor, GradientBoostingRegressor, RandomForestRegressor
from sklearn.linear_model import LinearRegression, Ridge, SGDRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.svm import SVR, LinearSVR, NuSVR
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor


def get_dict_models():
    dict_models = {
        "BaseKeras": BaseKeras(),
        "LSTMKeras": LSTMKeras(),
        "XGBRegressor": xgb.XGBRegressor(),
        "RandomForestRegressor": RandomForestRegressor(),
        "GradientBoostingRegressor": GradientBoostingRegressor(),
        "AdaBoostRegressor": AdaBoostRegressor(),
        "LinearRegression": LinearRegression(),
        "Ridge": Ridge(),
        "SGDRegressor": SGDRegressor(),
        "LinearSVR": LinearSVR(),
        "SVR": SVR(),
        "NuSVR": NuSVR(),
        "DecisionTreeRegressor": DecisionTreeRegressor(),
        "ExtraTreeRegressor": ExtraTreeRegressor(),
        "MLPRegressor": MLPRegressor(),
        "KNeighborsRegressor": KNeighborsRegressor(),
        "ARMA": ARMA(),
        #  'BaselineLastYear': BaselineLastYear(),
        "BaselineLastValue": BaselineLastValue(),
        "BaselineMean": BaselineMean(),
        "BaselineMedian": BaselineMedian(),
        "Prophet": Prophet(),
    }
    return dict_models



"""
utils_coinbase.py
"""
import asyncio
import sys
import time
from datetime import datetime, timedelta
from typing import List

import pandas as pd
import requests
from aiohttp import ClientSession
from autoforecast.src.utils.logger import LOG
from tqdm import tqdm


async def _get_current_price(type="buy", currency_pair="BTC-USD", date=None, session=None):
    if date is None:
        date = datetime.now().strftime("%Y-%m-%d")
    available_type = ["buy", "sell", "spot"]
    BASE_URL = "https://api.coinbase.com/v2/prices/"
    if currency_pair is not None:
        url = "{}{}/".format(BASE_URL, currency_pair)
    else:
        LOG.error("Error: currency_pair is None")
        sys.exit(1)
    if type in available_type:
        url = "{}{}".format(url, type)
    else:
        LOG.error("Error: type not available, try {}".format(available_type))
        sys.exit(1)
    if date is not None:
        url = "{}?date={}".format(url, date)
    try:
        async with session.get(url) as response:
            if response.status != 200:
                LOG.error("status_code {} for url {}".format(response.status, url))
            else:
                json = response.json()
                return await json
    except Exception as e:
        LOG.error("Not able to get {}, {}".format(url, e))
        return {}


def _get_current_price_loop(type="buy", currency_pair="BTC-USD", date=None, session=None):
    if date is None:
        date = datetime.now().strftime("%Y-%m-%d")
    available_type = ["buy", "sell", "spot"]
    BASE_URL = "https://api.coinbase.com/v2/prices/"
    if currency_pair is not None:
        url = "{}{}/".format(BASE_URL, currency_pair)
    else:
        LOG.error("Error: currency_pair is None")
        sys.exit(1)
    if type in available_type:
        url = "{}{}".format(url, type)
    else:
        LOG.error("Error: type not available, try {}".format(available_type))
        sys.exit(1)
    if date is not None:
        url = "{}?date={}".format(url, date)
    try:
        response = requests.get(url)
        if response.status_code != 200:
            LOG.error("status_code {} for url {}".format(response.status_code, url))
        else:
            json = response.json()
            return json
    except Exception as e:
        LOG.error("Not able to get {}, {}".format(url, e))
        return {}


async def future_price(dates: List[str], type, currency_pair) -> object:
    """
    Async function to do API call faster.
    input:
            :mdls: list, models we want ratings
            :token: str, openvoice token to query the API
            :endpoint: str, endpoint from OPV API
    output:
            asyncio object
    """
    tasks = []
    async with ClientSession() as session:
        for date in dates:
            task = asyncio.ensure_future(
                _get_current_price(
                    type=type, currency_pair=currency_pair, date=date, session=session
                )
            )
            tasks.append(task)

        return await asyncio.gather(*tasks)


def get_price_for_last_n_days(n=1, type="spot", currency_pair="BTC-USD"):
    LOG.debug("Fetching {} ({}) data for the last {} days...".format(type, currency_pair, n))
    start = time.time()
    dates = []
    prices = []
    jsons = []
    # loop = asyncio.get_event_loop()
    for i in tqdm(range(n, -1, -1)):
        date = (datetime.now() - timedelta(days=i)).strftime("%Y-%m-%d")
        dates.append(date)
        jsons.append(_get_current_price_loop(type=type, currency_pair=currency_pair, date=date))
    """
    future = asyncio.ensure_future(future_price(dates=dates,
                                                type=type,
                                                currency_pair=currency_pair))
    jsons = loop.run_until_complete(future)
    loop.close()
    """
    prices = [float(json.get("data", {"amount": 0})["amount"]) if json else 0 for json in jsons]

    data = pd.DataFrame(
        {
            "date": dates,
            "price": prices,
            "timestamp": [i + 1 for i in range(len(dates))],
        }
    )
    LOG.debug("Data fetched in {:.2f} s".format(time.time() - start))
    return data



from typing import Callable

from skopt import gp_minimize
from skopt.utils import use_named_args


class HyperparametersTuner:
    def __init__(
        self,
        model: Callable,
        search_space: list,
        x0: list,
        metric: Callable,
        X_train,
        X_test,
        y_train,
        y_test,
    ):
        self.model = model
        self.search_space = search_space
        self.x0 = x0
        self.metric = metric
        self.X_train = X_train
        self.X_test = X_test
        self.y_train = y_train
        self.y_test = y_test

    def __call__(self):
        return self.optimize()

    def optimize(
        self, n_calls: int = 10, n_random_starts: int = 3, random_state: int = 666
    ):
        dimensions = self.search_space
        print(dimensions)

        @use_named_args(dimensions=dimensions)
        def fitness(**params):
            print(f"params={params}")
            model = self.model()
            model.fit(self.X_train, self.y_train, **params)

            y_pred = model.predict(self.X_test)
            metric_value = self.metric(self.y_test, y_pred)
            print(f"metric_value={metric_value}")
            return metric_value

        res = gp_minimize(
            func=fitness,
            dimensions=dimensions,
            acq_func="EI",  # Expected Improvement.
            x0=self.x0,
            n_calls=n_calls,
            n_random_starts=n_random_starts,
            random_state=random_state,
        )
        print(f"best accuracy={-1.0 * res.fun} with {res.x}")
        param_name = [space.name for space in self.search_space]
        best_params = dict(zip(param_name, res.x))
        return res, best_params



from sklearn.preprocessing import MinMaxScaler


def scale_transform(train, test=None):
    scaler = MinMaxScaler()
    scaler.fit(train)
    train = scaler.transform(train)
    if test is not None:
        test = scaler.transform(test)
    return train, test, scaler


def run(df, list_num_feat):
    train, test, scaler = scale_transform(train=df[list_num_feat])
    return train, test, scaler



from typing import List, Optional

from autoforecast.preprocessing import categorical, engineering, features_selection, numerical


def preprocessing(
    df,
    target_name: List[str],
    categoricals: List[str],
    numericals: List[str],
    date_col: Optional[str] = None,
    train_size: float = 0.7,
    engineering: bool = True,
    selection: bool = True,
):
    cutoff = int(len(df) * train_size)
    train, test = df[:cutoff], df[cutoff:]
    X_train, y_train, X_test, y_test = None
    return X_train, y_train, X_test, y_test



from autoforecast import metrics
from autoforecast.models.hyperparameters import HyperparametersTuner


class BaseMLRegressor:
    def __init__(self):
        pass

    def fit(self):
        self.model.fit()

    def predic(self):
        self.model.predict()

    def optimize(self, X_train, X_test, y_train, y_test):
        return HyperparametersTuner(
            model=None,
            search_space=None,
            x0=None,
            metric=metrics.smape_score,
            X_train=X_train,
            X_test=X_test,
            y_train=y_train,
            y_test=y_test,
        )()



import zipfile

import pandas as pd
import requests
from autoforecast.src.utils.logger import LOG


def download_url(url, save_path, chunk_size=128):
    r = requests.get(url, stream=True)
    with open(save_path, "wb") as fd:
        for chunk in r.iter_content(chunk_size=chunk_size):
            fd.write(chunk)


def import_ml_db(data="energy"):

    if data == "air_quality":
        file_url = (
            "https://archive.ics.uci.edu/ml/machine-learning-databases/00360/AirQualityUCI.zip"
        )
        download_url(url=file_url, save_path="AirQualityUCI", chunk_size=128)
        path_to_zip_file = "AirQualityUCI"
        directory_to_extract_to = "temp_dir"
        with zipfile.ZipFile(path_to_zip_file, "r") as zip_ref:
            zip_ref.extractall(directory_to_extract_to)
        df = pd.read_csv("temp_dir/AirQualityUCI.csv", sep=";")
        LOG.debug(df.head())

    elif data == "energy":
        file_url = "https://archive.ics.uci.edu/ml/machine-learning-databases/00374/energydata_complete.csv"
        df = pd.read_csv(file_url)
        LOG.debug(df.head())
    return df
